//
//load_elf.c
//- a crude .elf file loader
//-see http://www.skyfree.org/linux/references/ELF_Format.pdf
//

#include <tap.h>

#define SEEK_SET 0
#define to_printable(c) ((c>0x20) && (c<0x80) ? c : ' ')

#define EI_NIDENT 16
typedef struct{
    dword   p_type;
    dword   p_offset;
    dword   p_vaddr;
    dword   p_paddr;
    dword   p_filesz;
    dword   p_memsz;
    dword   p_flags;
    dword   p_align;
}Elf32_Phdr;

typedef struct{
    byte    e_ident[EI_NIDENT];
    word    e_type;
    word    e_machine;
    dword   e_version;
    dword   e_entry;
    dword   e_phoff;
    dword   e_shoff;
    dword   e_flags;
    word    e_ehsize;
    word    e_phentsize;
    word    e_phnum;
    word    e_shentsize;
    word    e_shnum;
    word    e_shstrndx;
    Elf32_Phdr p[10];
}Elf32_Ehdr;

typedef struct{
    dword   sh_name;
    dword   sh_type;
    dword   sh_flags;
    dword   sh_addr;
    dword   sh_offset;
    dword   sh_size;
    dword   sh_link;
    dword   sh_info;
    dword   sh_addralign;
    dword   sh_entsize;
}Elf32_Shdr;

typedef struct{
    dword   st_name;
    dword   st_value;
    dword   st_size;
    dword   st_info;
    dword   st_other;
    word    st_shndx;
}Elf32_Sym;

enum {
    SHT_NULL     = 0,
    SHT_PROGBITS = 1,
    SHT_SYMTAB   = 2,
    SHT_STRTAB   = 3,
    SHT_RELA     = 4,
    SHT_HASH     = 5,
    SHT_DYNAMIC  = 6,
    SHT_NOTE     = 7,
    SHT_NOBITS   = 8,
    SHT_REL      = 9,
    SHT_SHLIB    = 10,
    SHT_DYNSYM   = 11,
    SHT_LOPROC   = 0x70000000,
    SHT_HIPROC   = 0x7fffffff,
    SHT_LOUSER   = 0x80000000,
    SHT_HIUSER   = 0xffffffff,
};

extern byte *fbuf;
extern int flen;
int debug = 0;

int load_elf(char *filename, dword *load, dword *entry)
{
    TYPE_File *Ip;
    Elf32_Ehdr *elf;
    Elf32_Shdr *s;
    int i;
    int ret = 0;
    byte *hbuf;
    byte *sbuf;
    byte *strtab;

    if (debug) TAP_Print("\nload_kernel(%s)", filename);
    TAP_Hdd_ChangeDir("/");
    TAP_Hdd_ChangeDir("ProgramFiles");
    Ip = TAP_Hdd_Fopen(filename);
    if (!Ip) {
        TAP_Print("\nfile open failed");
        return -1;
    }
    flen = TAP_Hdd_Flen(Ip);
    if (debug) TAP_Print("\nflen = %x", flen);

    hbuf = (byte *)malloc(0x400);
    if (!hbuf) {
        TAP_Print("\nmalloc failed");
        ret = -1;
        goto exit;
    }
    sbuf = (byte *)malloc(0x400);
    if (!sbuf) {
        TAP_Print("\nmalloc failed");
        ret = -1;
        goto exit;
    }

    //can't be longer than the entire .elf file
    fbuf = (byte *)malloc(flen);
    if (!fbuf) {
        TAP_Print("\nmalloc failed");
        ret = -1;
        goto exit;
    }
    //initialise to ensure holes are zero
    memset(&fbuf[0], 0, flen);

    //load the header
    TAP_Hdd_Fseek(Ip, 0, SEEK_SET);
    TAP_Hdd_Fread(&hbuf[0], 0x400, 1, Ip);
    elf = (Elf32_Ehdr *)&hbuf[0];
    if (debug) {
        TAP_Print("\ne_ident-magic   = %08x %c%c%c%c", *(dword *)(&elf->e_ident[0]), to_printable(elf->e_ident[0]), to_printable(elf->e_ident[1]), to_printable(elf->e_ident[2]), to_printable(elf->e_ident[3]));
        TAP_Print("\ne_ident-class   = %02x", elf->e_ident[4]);
        TAP_Print("\ne_ident-data    = %02x", elf->e_ident[5]);
        TAP_Print("\ne_ident-version = %02x", elf->e_ident[6]);
        TAP_Print("\ne_type          = %04x", elf->e_type);
        TAP_Print("\ne_machine       = %04x", elf->e_machine);
        TAP_Print("\ne_version       = %08x", elf->e_version);
        TAP_Print("\ne_entry         = %08x", elf->e_entry);
        TAP_Print("\ne_phoff         = %08x", elf->e_phoff);
        TAP_Print("\ne_shoff         = %08x", elf->e_shoff);
        TAP_Print("\ne_flags         = %08x", elf->e_flags);
        TAP_Print("\ne_ehsize        = %04x", elf->e_ehsize);
        TAP_Print("\ne_phentsize     = %04x", elf->e_phentsize);
        TAP_Print("\ne_phnum         = %04x", elf->e_phnum);
        TAP_Print("\ne_shentsize     = %04x", elf->e_shentsize);
        TAP_Print("\ne_shnum         = %04x", elf->e_shnum);
        TAP_Print("\ne_shstrndx      = %04x", elf->e_shstrndx);
    }

    if ((elf->e_ident[0] != 0x7f) && (elf->e_ident[1] != 'E') && (elf->e_ident[2] != 'L') && (elf->e_ident[3] != 'F')) {
        TAP_Print("\nnot an elf file");
        ret = -1;
        goto exit;
    }

    if (elf->e_ident[4] != 0x01) {
        TAP_Print("\nnot a 32 bit class");
        ret = -1;
        goto exit;
    }

    if (elf->e_ident[5] != 0x02) {
        TAP_Print("\nnot in big endian format");
        ret = -1;
        goto exit;
    }

    if (elf->e_ident[6] != 0x01) {
        TAP_Print("\nnot a valid version");
        ret = -1;
        goto exit;
    }

    if (elf->e_machine != 8) {
        TAP_Print("\nnot a mips file");
        ret = -1;
        goto exit;
    }

    if (elf->e_version != 1) {
        TAP_Print("\nnot a current version");
        ret = -1;
        goto exit;
    }

    if (debug) {
        for (i=0; i<elf->e_phnum; i++) {
            TAP_Print("\n");
            TAP_Print("\nprogram header %d, %x", i, elf->e_phoff+i*elf->e_phentsize);
            TAP_Print("\np_type       = %08x", elf->p[i].p_type);
            TAP_Print("\np_offset     = %08x", elf->p[i].p_offset);
            TAP_Print("\np_vaddr      = %08x", elf->p[i].p_vaddr);
            TAP_Print("\np_paddr      = %08x", elf->p[i].p_paddr);
            TAP_Print("\np_filesz     = %08x", elf->p[i].p_filesz);
            TAP_Print("\np_memsz      = %08x", elf->p[i].p_memsz);
            TAP_Print("\np_flags      = %08x", elf->p[i].p_flags);
            TAP_Print("\np_align      = %08x", elf->p[i].p_align);
        }
    }


    //load the elf header
    TAP_Hdd_Fseek(Ip, elf->e_shoff, SEEK_SET);
    TAP_Hdd_Fread(&sbuf[0], 0x400, 1, Ip);


    //get the string table
    s = (Elf32_Shdr *)&sbuf[elf->e_shstrndx*elf->e_shentsize];
    if (s->sh_type == SHT_STRTAB) {
        if (debug) TAP_Print("\nread string table");
        strtab = (byte *)malloc(s->sh_size);
        ret = TAP_Hdd_Fseek(Ip, s->sh_offset, SEEK_SET);
        ret = TAP_Hdd_Fread(&strtab[0], s->sh_size, 1, Ip);
        if (debug) dump(&strtab[0], s->sh_size);
    }


    //now trawl through the sections
    for (i=0; i<elf->e_shnum; i++) {
        s = (Elf32_Shdr *)&sbuf[i*elf->e_shentsize];
        if (debug) {
            TAP_Print("\n");
            TAP_Print("\nsection header %d, %x", i, i*elf->e_shentsize);
            TAP_Print("\nsh_name      = %08x, %s", s->sh_name, &strtab[s->sh_name]);
            TAP_Print("\nsh_type      = %08x", s->sh_type);
            TAP_Print("\nsh_flags     = %08x", s->sh_flags);
            TAP_Print("\nsh_addr      = %08x", s->sh_addr);
            TAP_Print("\nsh_offset    = %08x", s->sh_offset);
            TAP_Print("\nsh_size      = %08x", s->sh_size);
            TAP_Print("\nsh_link      = %08x", s->sh_link);
            TAP_Print("\nsh_info      = %08x", s->sh_info);
            TAP_Print("\nsh_addralign = %08x", s->sh_addralign);
            TAP_Print("\nsh_entsize   = %08x", s->sh_entsize);
        }

        //ignore sections that are not SHT_PROGBITS, are zero length, or are below our baseaddress
        if ((s->sh_type == SHT_PROGBITS)  && (s->sh_size > 0) && (s->sh_addr >= elf->p[0].p_paddr)) {
            TAP_Print("\nload file    : %08x+%6x bytes to mem: %08x+%6x", s->sh_offset, s->sh_size, s->sh_addr, s->sh_size);

            ret = TAP_Hdd_Fseek(Ip, s->sh_offset, SEEK_SET);
            ret = TAP_Hdd_Fread(&fbuf[s->sh_addr-elf->p[0].p_paddr], s->sh_size, 1, Ip);
            if (ret == 0) {
                TAP_Print("\nTAP_Hdd_Fread failed");
                free(strtab);
                ret = -1;
                goto exit;
            }

            //the total length is the offset+length of the final section
            flen = s->sh_addr-elf->p[0].p_paddr+s->sh_size;
        }
    }
    free(strtab);

    *load = elf->p[0].p_paddr;      //<kernel_load>
    *entry = elf->e_entry;          //<kernel_entry>


exit:
    free(hbuf);
    free(sbuf);
    TAP_Hdd_Fclose(Ip);
    return ret;
}
